contracts = {
    "vana": {
        "TeePool": "0x3c92fD91639b41f13338CE62f19131e7d19eaa0D",
        "DataRegistry": "0x8C8788f98385F6ba1adD4234e551ABba0f82Cb7C",
        "RootNetworkContract": "0xff14346dF2B8Fd0c95BF34f1c92e49417b508AD5",
        "DataRefinerRegistry": "0x93c3EF89369fDcf08Be159D9DeF0F18AB6Be008c",
        "ComputeInstructionRegistry": "0x5786B12b4c6Ba2bFAF0e77Ed30Bf6d32805563A5",
        "ComputeEngine": "0xb2BFe33FA420c45F1Cf1287542ad81ae935447bd",
        "QueryEngine": "0xd25Eb66EA2452cf3238A2eC6C1FD1B7F5B320490",
    },
    "islander": {
        "TeePool": "0x3c92fD91639b41f13338CE62f19131e7d19eaa0D",
        "DataRegistry": "0x8C8788f98385F6ba1adD4234e551ABba0f82Cb7C",
        "RootNetworkContract": "0xff14346dF2B8Fd0c95BF34f1c92e49417b508AD5",
        "DataRefinerRegistry": "0x93c3EF89369fDcf08Be159D9DeF0F18AB6Be008c",
        "ComputeInstructionRegistry": "0x5786B12b4c6Ba2bFAF0e77Ed30Bf6d32805563A5",
        "ComputeEngine": "0xb2BFe33FA420c45F1Cf1287542ad81ae935447bd",
        "QueryEngine": "0xd25Eb66EA2452cf3238A2eC6C1FD1B7F5B320490",
    },
    "maya": {
        "TeePool": "0xF084Ca24B4E29Aa843898e0B12c465fAFD089965",
        "DataRegistry": "0xEA882bb75C54DE9A08bC46b46c396727B4BFe9a5",
        "RootNetworkContract": "0xf408A064d640b620219F510963646Ed2bD5606BB",
        "DataRefinerRegistry": "0x93c3EF89369fDcf08Be159D9DeF0F18AB6Be008c",
        "ComputeInstructionRegistry": "0x5786B12b4c6Ba2bFAF0e77Ed30Bf6d32805563A5",
        "ComputeEngine": "0xb2BFe33FA420c45F1Cf1287542ad81ae935447bd",
        "QueryEngine": "0xd25Eb66EA2452cf3238A2eC6C1FD1B7F5B320490",
    },
    "satori": {
        "TeePool": "0xF084Ca24B4E29Aa843898e0B12c465fAFD089965",
        "DataRegistry": "0xEA882bb75C54DE9A08bC46b46c396727B4BFe9a5",
        "RootNetworkContract": "0xf408A064d640b620219F510963646Ed2bD5606BB",
        "DataRefinerRegistry": "0x93c3EF89369fDcf08Be159D9DeF0F18AB6Be008c",
        "ComputeInstructionRegistry": "0x5786B12b4c6Ba2bFAF0e77Ed30Bf6d32805563A5",
        "ComputeEngine": "0xb2BFe33FA420c45F1Cf1287542ad81ae935447bd",
        "QueryEngine": "0xd25Eb66EA2452cf3238A2eC6C1FD1B7F5B320490",
    },
    "moksha": {
        "TeePool": "0x3c92fD91639b41f13338CE62f19131e7d19eaa0D",
        "DataRegistry": "0x8C8788f98385F6ba1adD4234e551ABba0f82Cb7C",
        "RootNetworkContract": "0xff14346dF2B8Fd0c95BF34f1c92e49417b508AD5",
        "DataRefinerRegistry": "0x93c3EF89369fDcf08Be159D9DeF0F18AB6Be008c",
        "ComputeInstructionRegistry": "0x5786B12b4c6Ba2bFAF0e77Ed30Bf6d32805563A5",
        "ComputeEngine": "0xb2BFe33FA420c45F1Cf1287542ad81ae935447bd",
        "QueryEngine": "0xd25Eb66EA2452cf3238A2eC6C1FD1B7F5B320490",
    },
}
